/*
->================================================================================<-
->= JoyDivision - USB joystick adapter - (c) Copyright 2016-2019 OnyxSoft        =<-
->================================================================================<-
->= Version  : 0.4                                                               =<-
->= File     : sega.c                                                            =<-
->= Author   : Stefan Blixth (stefan@onyxsoft.se)                                =<-
->= Compiled : 2019-08-15                                                        =<-
->================================================================================<-
->=                                                                              =<-
->= This file is part of JoyDivision - USB joystick adapter                      =<-
->=                                                                              =<-
->= JoyDivision is free software; you can redistribute it and/or modify          =<-
->= it under the terms of the GNU General Public License as published by         =<-
->= the Free Software Foundation; either version 2 of the License, or            =<-
->= (at your option) any later version.                                          =<-
->=                                                                              =<-
->= This program is distributed in the hope that it will be useful,              =<-
->= but WITHOUT ANY WARRANTY; without even the implied warranty of               =<-
->= MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                =<-
->= GNU General Public License for more details.                                 =<-
->=                                                                              =<-
->= You should have received a copy of the GNU General Public License along      =<-
->= with this program; if not, write to the Free Software Foundation, Inc.,      =<-
->= 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.                  =<-
->=                                                                              =<-
->= This project is based on V-USB code from Objective Development Software GmbH =<-
->= http://www.obdev.at/products/vusb/                                           =<-
->=                                                                              =<-
->= Partial based on code/ideas from Raphal Assnat                             =<-
->= http://www.raphnet.net/                                                      =<-
->=                                                                              =<-
->= Partial based on code/ideas from Andreas Paul                                =<-
->= http://www.hexagons.de/                                                      =<-
->=                                                                              =<-
->================================================================================<-
*/

//Modifies by Hz for Arduino Nano V3 in 2023

#include "sega.h"

byte joy1[] = {5, 6, 7, 8, 9, 11, 10},
     joy2[] = {14, 15, 16, 17, 18, 19, 12};

unsigned long _lastReadTime = 0;
const unsigned long SC_READ_DELAY_MS = 5; // Must be >= 3 to give 6-button controller time to reset


#define JOYPORT1   0
#define JOYPORT2   1

#ifdef DUAL_JOYDIVISION

// USB configuration descriptor
PROGMEM const char usbDescriptorConfiguration[] = {
  9,                                       // sizeof(usbDescriptorConfiguration): length of descriptor in bytes
  USBDESCR_CONFIG,                         // descriptor type
  9 + (2*(9+9+7)), 0,                      // total length of data returned (including inlined descriptors)
  2,                                       // number of interfaces in this configuration
  1,                                       // index of this configuration
  0,                                       // configuration name string index
  USBATTR_BUSPOWER,                        // attributes
  USB_CFG_MAX_BUS_POWER/2,                 // max USB current in 2mA units

  // interface 1 descriptor follows inline:
  9,                                       // sizeof(usbDescrInterface): length of descriptor in bytes
  USBDESCR_INTERFACE,                      // descriptor type
  JOYPORT1,                                // index of this interface
  0,                                       // alternate setting for this interface
  USB_CFG_HAVE_INTRIN_ENDPOINT,            // endpoints excl 0: number of endpoint descriptors to follow
  USB_CFG_INTERFACE_CLASS,
  USB_CFG_INTERFACE_SUBCLASS,
  USB_CFG_INTERFACE_PROTOCOL,
  3,                                       // string index for interface
  9,                                       // sizeof(usbDescrHID): length of descriptor in bytes
  USBDESCR_HID,                            // descriptor type: HID
  0x01, 0x01,                              // BCD representation of HID version
  0x00,                                    // target country code
  0x01,                                    // number of HID Report (or other HID class) Descriptor infos to follow
  0x22,                                    // descriptor type: report
  USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH, 0, // total length of report descriptor

  // endpoint descriptor for endpoint 1
  7,                                       // sizeof(usbDescrEndpoint)
  USBDESCR_ENDPOINT,                       // descriptor type = endpoint
  (char)0x81,                              // IN endpoint number 1
  0x03,                                    // attrib: Interrupt endpoint
  8, 0,                                    // maximum packet size
  USB_CFG_INTR_POLL_INTERVAL,              // in ms

  // interface 2 descriptor follows inline:
  9,                                       // sizeof(usbDescrInterface): length of descriptor in bytes
  USBDESCR_INTERFACE,                      // descriptor type
  JOYPORT2,                                // index of this interface
  0,                                       // alternate setting for this interface
  1,                                       // endpoints excl 0: number of endpoint descriptors to follow
  USB_CFG_INTERFACE_CLASS,
  USB_CFG_INTERFACE_SUBCLASS,
  USB_CFG_INTERFACE_PROTOCOL,
  4,                                       // string index for interface
  9,                                       // sizeof(usbDescrHID): length of descriptor in bytes
  USBDESCR_HID,                            // descriptor type: HID
  0x01, 0x01,                              // BCD representation of HID version
  0x00,                                    // target country code
  0x01,                                    // number of HID Report (or other HID class) Descriptor infos to follow
  0x22,                                    // descriptor type: report
  USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH, 0, // total length of report descriptor

  // endpoint descriptor for endpoint 3
  7,                                       // sizeof(usbDescrEndpoint)
  USBDESCR_ENDPOINT,                       // descriptor type = endpoint
  (char)(0x80 | USB_CFG_EP3_NUMBER),       // IN endpoint number 3
  0x03,                                    // attrib: Interrupt endpoint
  8, 0,                                    // maximum packet size
  USB_CFG_INTR_POLL_INTERVAL,              // in ms
};

#endif


PROGMEM const char usbHidReportDescriptor[]  =
{
    // Joystick
    0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
    0x09, 0x05,                    // USAGE (Game Pad)
    0xa1, 0x01,                    // COLLECTION (Application)
      0xa1, 0x02,                  //   COLLECTION (Physical)

         // Buttons
         0x05, 0x09,               //     USAGE_PAGE (Button)
         0x19, 0x01,               //     USAGE_MINIMUM (Button 1)
         0x29, 0x08,               //     USAGE_MAXIMUM (Button 8)
         0x15, 0x00,               //     LOGICAL_MINIMUM (0)
         0x25, 0x01,               //     LOGICAL_MAXIMUM (1)
         0x95, 0x08,               //     REPORT_COUNT (8)
         0x75, 0x01,               //     REPORT_SIZE (1)
         0x81, 0x02,               //     INPUT (Data,Var,Abs)
         0x95, 0x00,               //     REPORT_COUNT (0)
         0x75, 0x01,               //     REPORT_SIZE (1)
         0x81, 0x03,               //     INPUT (Constant,Var,Abs)

         // Axis
         0x05, 0x01,               //     USAGE_PAGE (Generic Desktop)
         0x09, 0x30,               //     USAGE (X)
         0x09, 0x31,               //     USAGE (Y)
         0x15, 0x81,               //     LOGICAL_MINIMUM (-127)
         0x25, 0x7f,               //     LOGICAL_MAXIMUM (127)
         0x75, 0x08,               //     REPORT_SIZE (8)
         0x95, 0x02,               //     REPORT_COUNT (2)
         0x81, 0x02,               //     INPUT (Data,Var,Abs)

      0xc0,                        //   END_COLLECTION
    0xc0                           // END_COLLECTION
};

typedef struct
{
   uchar buttons;
   char  axis_x;
   char  axis_y;
}report_t;


#ifdef DUAL_JOYDIVISION
static report_t reportBuffer[2];
static PROGMEM const char DeviceStr[] = "Dual JoyDivision #x";
static PROGMEM const char VendorStr[] = "OnyxSoft";
static const char Desc0Str[] = { 4, 3, 0x09, 0x05 };

usbMsgLen_t usbFunctionDescriptor(struct usbRequest *rq)
{
  static char buf[(sizeof(DeviceStr) > sizeof(VendorStr) ? sizeof(DeviceStr) : sizeof(VendorStr))*2];

  uchar ix = rq->wValue.bytes[0];
  const char* s;
  uchar l;
  uchar ifchar = 0;
  if (ix == 0) { // String0
    usbMsgPtr = (usbMsgPtr_t)Desc0Str;
    return 4;
  }
  else if (ix == 1) { // Vendor
    s = VendorStr;
    l = sizeof(VendorStr) * 2;
  }
  else if (ix == 2) { // Product
    s = DeviceStr;
    l = (sizeof(DeviceStr) - 3) * 2;
  }
  else if (ix == 3 || ix == 4) { // Interface 1 or 2
    s = DeviceStr;
    l = sizeof(DeviceStr) * 2;
    ifchar = '1' + ix - 3;
  }
  else {
    return 0;
  }

  // copy from ROM to RAM with char to wchar conversion
  char* d = buf + 2;
  char c;
  while ((c = pgm_read_byte(s++)) != 0) {
    *d = c;
    d += 2;
  }

  buf[0] = l; // length
  buf[1] = 3; // string descriptor mark
  if (ifchar)
    buf[(sizeof(DeviceStr)-1) * 2] = ifchar; // patch interface number
  usbMsgPtr = (usbMsgPtr_t)buf;
  return l;
}

#else
static report_t reportBuffer;
#endif

static uchar idleRate = 0;            // repeat rate for keyboards, never used for mice

enum {
      NORMAL,
      SWITCH,
      JOY1ONLY,
      JOY2ONLY,
      COPYFROM1,
      COPYFROM2
} inputMode;

int lastMode = NORMAL;

static report_t lastBuffer[2] = {};

int input(joyport){
      switch (inputMode)
      {
        case SWITCH:
          return ((joyport == 0)?1 : 0);
        case COPYFROM1:
          return 0;
        case COPYFROM2:
          return 1;
        default:
          return joyport;          
      }
}

void switchInput(char joyport, char buttons){

      if ((buttons & 0x08) && !(lastBuffer[joyport].buttons & 0x08))
            if (inputMode == SWITCH) inputMode = NORMAL;
            else inputMode = SWITCH;
      if ((buttons & 0x10) && !(lastBuffer[joyport].buttons & 0x10))
            if (inputMode == COPYFROM1) inputMode = NORMAL;
            else inputMode = COPYFROM1;
      if ((buttons & 0x20) && !(lastBuffer[joyport].buttons & 0x20))
            if (inputMode == COPYFROM2) inputMode = NORMAL;
            else inputMode = COPYFROM2;
}

//report_t buildReport(uchar joyport)
void buildReport(uchar joyport)
{
   char axis_x=0, axis_y=0, sixButtonMode =  0;
   uchar buttons=0;
   uchar *joy = (joyport == 0? joy1 : joy2);
   report_t *activeBuffer;

 #ifdef DUAL_JOYDIVISION
   activeBuffer = &reportBuffer[joyport];
#else
   activeBuffer = &reportBuffer;
#endif

      digitalWrite(joy[6],LOW); // Set SELECT pin low...
      _delay_us(20);
      digitalWrite(joy[6],HIGH); // Set SELECT pin high...
      _delay_us(20);

      digitalWrite(joy[6],LOW); // Set SELECT pin low...
      _delay_us(20);

      if ((digitalRead(joy[2]) == LOW) && (digitalRead(joy[3]) == LOW)){
            if (digitalRead(joy[4]) == LOW) buttons |= 0x01; // Check if Button A is active
            if (digitalRead(joy[5]) == LOW)buttons |= 0x40; // Check if Start button  is active
      }
   
      digitalWrite(joy[6],HIGH); // Set SELECT pin high...
      _delay_us(20);
   
      if (digitalRead(joy[0]) == LOW) axis_y = 0x81;   // Check if Up-axis is active
      if (digitalRead(joy[1]) == LOW) axis_y = 0x7f;   // Check if Down-axis is active
      if (digitalRead(joy[2]) == LOW) axis_x = 0x81;   // Check if Left-axis is active
      if (digitalRead(joy[3]) == LOW) axis_x = 0x7f;   // Check if Right-axis is active
      if (digitalRead(joy[4]) == LOW) buttons |= 0x02; // Check if Button B is active
      if (digitalRead(joy[5]) == LOW) buttons |= 0x04; // Check if Button C is active

      //CLEARBIT(PORTB, PB1); // Set SELECT pin low...
      digitalWrite(joy[6],LOW);
      _delay_us(20);

     sixButtonMode = (digitalRead(joy[0]) == LOW) && (digitalRead(joy[1]) == LOW);

     digitalWrite(joy[6],HIGH); // Set SELECT pin high...
      _delay_us(20);

      if (sixButtonMode)
      {
               if (digitalRead(joy[3]) == LOW) buttons |= 0x80; // Check if Mode is active
               if (digitalRead(joy[0]) == LOW) buttons |= 0x20; // Check if Button Z is active
               if (digitalRead(joy[1]) == LOW) buttons |= 0x10; // Check if Button Y is active
               if (digitalRead(joy[2]) == LOW) buttons |= 0x08; // Check if Button X is active 
               if ((buttons & 0x40 ) && (buttons & 0x80)) //if mode + start
               switchInput(joyport, buttons);
      }

   activeBuffer->axis_x  = axis_x;
   activeBuffer->axis_y  = axis_y;
   activeBuffer->buttons = buttons;
   lastBuffer[joyport] = *activeBuffer;

   _lastReadTime = millis();

   //return activeBuffer;
}

uchar usbFunctionSetup(uchar data[8])
{
   usbRequest_t *rq = (void *)data;

   if((rq->bmRequestType & USBRQ_TYPE_MASK) == USBRQ_TYPE_CLASS)
   {
      if(rq->bRequest == USBRQ_HID_GET_REPORT)
      {
         usbMsgPtr = (usbMsgPtr_t )&reportBuffer;
         return sizeof(reportBuffer);
      }
      else if(rq->bRequest == USBRQ_HID_GET_IDLE)
      {
         usbMsgPtr = (usbMsgPtr_t)&idleRate;
         return 1;
      }
      else if(rq->bRequest == USBRQ_HID_SET_IDLE)
      {
         idleRate = rq->wValue.bytes[1];
         return 0;
      }
   }

   return 0;
}


static void hardwareInit(void)
{
   uchar i, j; // Test for USB Reset

   /*      1 _____________ 5
             \ o o o o o /
              \ o o o o /
            6   9


   Sega info borrowed from : http://segaretro.org/Six_Button_Control_Pad_(Mega_Drive)
   http://www.cs.cmu.edu/~chuck/infopg/segasix.txt

   Technical Information

   In order to maintain backwards-compatibility with existing games, the 6-button gamepad acts the same as the 3-button if the 
   controller inputs are read by the program a maximum of four times per frame. Reading the 3-button gamepad takes two cycles 
   (once with TH=1 to read Up, Down, Left, Right, B, and C; and once with TH=0 to read A and Start), and the 6-button gamepad 
   allows this to occur twice. It then reports the state of the extra buttons when the TH line is changed after the fourth cycle.

   The full controller state is read in 8 cycles:

   Cycle    TH out   TR in    TL in    D3 in    D2 in    D1 in    D0 in
   1        HI       C        B        Right    Left     Down     Up
   2        LO       Start    A        0        0        Down     Up
   3        HI       C        B        Right    Left     Down     Up
   4        LO       Start    A        0        0        Down     Up
   5        HI       C        B        Right    Left     Down     Up
   6        LO       Start    A        0        0        0        0
   7        HI       C        B        Mode     X        Y        Z
   8        LO       Start    A        ---      ---      ---      ---

   Note: For the 8th cycle, '---' indicates "don't care". These values are all 1 on the 6-button controller, but may be mapped to 
   other buttons on third-party controllers. Software expecting 6-button controllers should simply ignore these values. (6-button 
   detection should be done on the 6th cycle, which presents four 0's on the D lines.)

   The controller's internal counter automatically resets after the eighth cycle. It also resets if it doesn't detect a rising-
   edge (0 to 1) transition on the TH line within 1.5 ms, which ensures that most games that don't support the extra buttons won't 
   read them accidentally.

   Some games incorrectly read the controller more than twice per frame, which may cause them to read the wrong set of buttons. If 
   the Mode button is held in while starting the Mega Drive console, the 6-button gamepad will disable the extra buttons and act   
   identically to a 3-button gamepad. (That is, the controller's internal counter will reset after the first two cycles instead of 
   after eight cycles.) 

----

   This means that we need to "clock" the states for at least 8 "rounds" to be able to read the gamepad states.

   Pin   Sega MD-GamePad     Arduino Nano-Pin (Port 1)   Arduino Nano-Pin (Port 2)
   1     Up / Z              5                14
   2     Down / Y            6                15
   3     Left / X            7                16
   4     Right / Mode        8                17
   5     +5V                 +5v              +5v
   6     B / A               9                18
   7     Select              11               12
   8     GND                 GND              GND
   9     C / Start           10               19  */

   pinMode(joy1[6], OUTPUT);
   digitalWrite(joy1[6], HIGH);
   pinMode(joy2[6], OUTPUT);
   digitalWrite(joy2[6], HIGH);

    // Setup input pins
    for (byte i = 0; i < 6; i++)
    {
        pinMode(joy1[i], INPUT_PULLUP);
    }

    for (byte i = 0; i < 6; i++)
    {
        pinMode(joy2[i], INPUT_PULLUP);
    }
    cli();
      usbDeviceDisconnect();
      _delay_ms(250);
      usbDeviceConnect();
}

#ifdef __cplusplus
extern "C"{
#endif 
void start(){
     uchar i;
   report_t reportBufferPort[2];

   wdt_enable(WDTO_2S);
   hardwareInit();
   odDebugInit();

   usbInit();
   usbDeviceDisconnect();  // enforce re-enumeration, do this while interrupts are disabled!
   i = 0;

   while(--i)
   {                       // fake USB disconnect for > 250 ms
      wdt_reset();
      _delay_ms(1);
   }

   usbDeviceConnect();
   sei();

   for(;;)
   {
      wdt_reset();
      usbPoll();

      if (usbInterruptIsReady())
      {
         if (max(millis() - _lastReadTime, 0) >= SC_READ_DELAY_MS)
       {
         // Add check to only report when changes have been made...
        buildReport(JOYPORT1);
        buildReport(JOYPORT2);
         reportBufferPort[JOYPORT1] = reportBuffer[input(0)];
         reportBufferPort[JOYPORT2] = reportBuffer[input(1)];
         
         
       } 
       usbSetInterrupt((void *)&reportBufferPort[JOYPORT1], sizeof(report_t));

         #ifdef DUAL_JOYDIVISION
                  usbSetInterrupt3((void *)&reportBufferPort[JOYPORT2], sizeof(report_t));
         #endif
      }
   }
}
#ifdef __cplusplus
} // extern "C"
#endif